#!/usr/bin/env python
# coding: utf-8

from __future__ import unicode_literals

# Allow direct execution
import os
import sys
import unittest
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

import copy

from test.helper import FakeHDL, assertRegexpMatches
from haruhi_dl import HaruhiDL
from haruhi_dl.compat import compat_str, compat_urllib_error

# extractors must be imported directly from their files,
# __init__.py points to lazy_extractors, which have no additional attributes/methods
from haruhi_dl.extractor.youtube import YoutubeIE
from haruhi_dl.extractor.common import InfoExtractor
from haruhi_dl.postprocessor.common import PostProcessor
from haruhi_dl.utils import ExtractorError, match_filter_func

TEST_URL = 'http://localhost/sample.mp4'


class HDL(FakeHDL):
    def __init__(self, *args, **kwargs):
        super(HDL, self).__init__(*args, **kwargs)
        self.downloaded_info_dicts = []
        self.msgs = []

    def process_info(self, info_dict):
        self.downloaded_info_dicts.append(info_dict)

    def to_screen(self, msg):
        self.msgs.append(msg)


def _make_result(formats, **kwargs):
    res = {
        'formats': formats,
        'id': 'testid',
        'title': 'testttitle',
        'extractor': 'testex',
        'extractor_key': 'TestEx',
    }
    res.update(**kwargs)
    return res


class TestFormatSelection(unittest.TestCase):
    def test_prefer_free_formats(self):
        # Same resolution => download webm
        hdl = HDL()
        hdl.params['prefer_free_formats'] = True
        formats = [
            {'ext': 'webm', 'height': 460, 'url': TEST_URL},
            {'ext': 'mp4', 'height': 460, 'url': TEST_URL},
        ]
        info_dict = _make_result(formats)
        yie = YoutubeIE(hdl)
        yie._sort_formats(info_dict['formats'])
        hdl.process_ie_result(info_dict)
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['ext'], 'webm')

        # Different resolution => download best quality (mp4)
        hdl = HDL()
        hdl.params['prefer_free_formats'] = True
        formats = [
            {'ext': 'webm', 'height': 720, 'url': TEST_URL},
            {'ext': 'mp4', 'height': 1080, 'url': TEST_URL},
        ]
        info_dict['formats'] = formats
        yie = YoutubeIE(hdl)
        yie._sort_formats(info_dict['formats'])
        hdl.process_ie_result(info_dict)
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['ext'], 'mp4')

        # No prefer_free_formats => prefer mp4 and flv for greater compatibility
        hdl = HDL()
        hdl.params['prefer_free_formats'] = False
        formats = [
            {'ext': 'webm', 'height': 720, 'url': TEST_URL},
            {'ext': 'mp4', 'height': 720, 'url': TEST_URL},
            {'ext': 'flv', 'height': 720, 'url': TEST_URL},
        ]
        info_dict['formats'] = formats
        yie = YoutubeIE(hdl)
        yie._sort_formats(info_dict['formats'])
        hdl.process_ie_result(info_dict)
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['ext'], 'mp4')

        hdl = HDL()
        hdl.params['prefer_free_formats'] = False
        formats = [
            {'ext': 'flv', 'height': 720, 'url': TEST_URL},
            {'ext': 'webm', 'height': 720, 'url': TEST_URL},
        ]
        info_dict['formats'] = formats
        yie = YoutubeIE(hdl)
        yie._sort_formats(info_dict['formats'])
        hdl.process_ie_result(info_dict)
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['ext'], 'flv')

    def test_format_selection(self):
        formats = [
            {'format_id': '35', 'ext': 'mp4', 'preference': 1, 'url': TEST_URL},
            {'format_id': 'example-with-dashes', 'ext': 'webm', 'preference': 1, 'url': TEST_URL},
            {'format_id': '45', 'ext': 'webm', 'preference': 2, 'url': TEST_URL},
            {'format_id': '47', 'ext': 'webm', 'preference': 3, 'url': TEST_URL},
            {'format_id': '2', 'ext': 'flv', 'preference': 4, 'url': TEST_URL},
        ]
        info_dict = _make_result(formats)

        hdl = HDL({'format': '20/47'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], '47')

        hdl = HDL({'format': '20/71/worst'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], '35')

        hdl = HDL()
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], '2')

        hdl = HDL({'format': 'webm/mp4'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], '47')

        hdl = HDL({'format': '3gp/40/mp4'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], '35')

        hdl = HDL({'format': 'example-with-dashes'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'example-with-dashes')

    def test_format_selection_audio(self):
        formats = [
            {'format_id': 'audio-low', 'ext': 'webm', 'preference': 1, 'vcodec': 'none', 'url': TEST_URL},
            {'format_id': 'audio-mid', 'ext': 'webm', 'preference': 2, 'vcodec': 'none', 'url': TEST_URL},
            {'format_id': 'audio-high', 'ext': 'flv', 'preference': 3, 'vcodec': 'none', 'url': TEST_URL},
            {'format_id': 'vid', 'ext': 'mp4', 'preference': 4, 'url': TEST_URL},
        ]
        info_dict = _make_result(formats)

        hdl = HDL({'format': 'bestaudio'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'audio-high')

        hdl = HDL({'format': 'worstaudio'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'audio-low')

        formats = [
            {'format_id': 'vid-low', 'ext': 'mp4', 'preference': 1, 'url': TEST_URL},
            {'format_id': 'vid-high', 'ext': 'mp4', 'preference': 2, 'url': TEST_URL},
        ]
        info_dict = _make_result(formats)

        hdl = HDL({'format': 'bestaudio/worstaudio/best'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'vid-high')

    def test_format_selection_audio_exts(self):
        formats = [
            {'format_id': 'mp3-64', 'ext': 'mp3', 'abr': 64, 'url': 'http://_', 'vcodec': 'none'},
            {'format_id': 'ogg-64', 'ext': 'ogg', 'abr': 64, 'url': 'http://_', 'vcodec': 'none'},
            {'format_id': 'aac-64', 'ext': 'aac', 'abr': 64, 'url': 'http://_', 'vcodec': 'none'},
            {'format_id': 'mp3-32', 'ext': 'mp3', 'abr': 32, 'url': 'http://_', 'vcodec': 'none'},
            {'format_id': 'aac-32', 'ext': 'aac', 'abr': 32, 'url': 'http://_', 'vcodec': 'none'},
        ]

        info_dict = _make_result(formats)
        hdl = HDL({'format': 'best'})
        ie = YoutubeIE(hdl)
        ie._sort_formats(info_dict['formats'])
        hdl.process_ie_result(copy.deepcopy(info_dict))
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'aac-64')

        hdl = HDL({'format': 'mp3'})
        ie = YoutubeIE(hdl)
        ie._sort_formats(info_dict['formats'])
        hdl.process_ie_result(copy.deepcopy(info_dict))
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'mp3-64')

        hdl = HDL({'prefer_free_formats': True})
        ie = YoutubeIE(hdl)
        ie._sort_formats(info_dict['formats'])
        hdl.process_ie_result(copy.deepcopy(info_dict))
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'ogg-64')

    def test_format_selection_video(self):
        formats = [
            {'format_id': 'dash-video-low', 'ext': 'mp4', 'preference': 1, 'acodec': 'none', 'url': TEST_URL},
            {'format_id': 'dash-video-high', 'ext': 'mp4', 'preference': 2, 'acodec': 'none', 'url': TEST_URL},
            {'format_id': 'vid', 'ext': 'mp4', 'preference': 3, 'url': TEST_URL},
        ]
        info_dict = _make_result(formats)

        hdl = HDL({'format': 'bestvideo'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'dash-video-high')

        hdl = HDL({'format': 'worstvideo'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'dash-video-low')

        hdl = HDL({'format': 'bestvideo[format_id^=dash][format_id$=low]'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'dash-video-low')

        formats = [
            {'format_id': 'vid-vcodec-dot', 'ext': 'mp4', 'preference': 1, 'vcodec': 'avc1.123456', 'acodec': 'none', 'url': TEST_URL},
        ]
        info_dict = _make_result(formats)

        hdl = HDL({'format': 'bestvideo[vcodec=avc1.123456]'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'vid-vcodec-dot')

    def test_format_selection_string_ops(self):
        formats = [
            {'format_id': 'abc-cba', 'ext': 'mp4', 'url': TEST_URL},
            {'format_id': 'zxc-cxz', 'ext': 'webm', 'url': TEST_URL},
        ]
        info_dict = _make_result(formats)

        # equals (=)
        hdl = HDL({'format': '[format_id=abc-cba]'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'abc-cba')

        # does not equal (!=)
        hdl = HDL({'format': '[format_id!=abc-cba]'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'zxc-cxz')

        hdl = HDL({'format': '[format_id!=abc-cba][format_id!=zxc-cxz]'})
        self.assertRaises(ExtractorError, hdl.process_ie_result, info_dict.copy())

        # starts with (^=)
        hdl = HDL({'format': '[format_id^=abc]'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'abc-cba')

        # does not start with (!^=)
        hdl = HDL({'format': '[format_id!^=abc]'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'zxc-cxz')

        hdl = HDL({'format': '[format_id!^=abc][format_id!^=zxc]'})
        self.assertRaises(ExtractorError, hdl.process_ie_result, info_dict.copy())

        # ends with ($=)
        hdl = HDL({'format': '[format_id$=cba]'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'abc-cba')

        # does not end with (!$=)
        hdl = HDL({'format': '[format_id!$=cba]'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'zxc-cxz')

        hdl = HDL({'format': '[format_id!$=cba][format_id!$=cxz]'})
        self.assertRaises(ExtractorError, hdl.process_ie_result, info_dict.copy())

        # contains (*=)
        hdl = HDL({'format': '[format_id*=bc-cb]'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'abc-cba')

        # does not contain (!*=)
        hdl = HDL({'format': '[format_id!*=bc-cb]'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'zxc-cxz')

        hdl = HDL({'format': '[format_id!*=abc][format_id!*=zxc]'})
        self.assertRaises(ExtractorError, hdl.process_ie_result, info_dict.copy())

        hdl = HDL({'format': '[format_id!*=-]'})
        self.assertRaises(ExtractorError, hdl.process_ie_result, info_dict.copy())

    def test_youtube_format_selection(self):
        order = [
            '18',
            # Apple HTTP Live Streaming
            '96', '95', '94', '93', '92', '132', '151',
            # 3D
            '85', '84', '102', '83', '101', '82', '100',
            # Dash video
            '137', '248', '136', '247', '135', '246',
            '245', '244', '134', '243', '133', '242', '160',
            # Dash audio
            '141', '172', '140', '171', '139',
        ]

        def format_info(f_id):
            info = YoutubeIE._formats[f_id].copy()

            # XXX: In real cases InfoExtractor._parse_mpd_formats() fills up 'acodec'
            # and 'vcodec', while in tests such information is incomplete since
            # commit a6c2c24479e5f4827ceb06f64d855329c0a6f593
            # test_HaruhiDL.test_youtube_format_selection is broken without
            # this fix
            if 'acodec' in info and 'vcodec' not in info:
                info['vcodec'] = 'none'
            elif 'vcodec' in info and 'acodec' not in info:
                info['acodec'] = 'none'

            info['format_id'] = f_id
            info['url'] = 'url:' + f_id
            return info
        formats_order = [format_info(f_id) for f_id in order]

        info_dict = _make_result(list(formats_order), extractor='youtube')
        hdl = HDL({'format': 'bestvideo+bestaudio'})
        yie = YoutubeIE(hdl)
        yie._sort_formats(info_dict['formats'])
        hdl.process_ie_result(info_dict)
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], '137+141')
        self.assertEqual(downloaded['ext'], 'mp4')

        info_dict = _make_result(list(formats_order), extractor='youtube')
        hdl = HDL({'format': 'bestvideo[height>=999999]+bestaudio/best'})
        yie = YoutubeIE(hdl)
        yie._sort_formats(info_dict['formats'])
        hdl.process_ie_result(info_dict)
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], '18')

        info_dict = _make_result(list(formats_order), extractor='youtube')
        hdl = HDL({'format': 'bestvideo/best,bestaudio'})
        yie = YoutubeIE(hdl)
        yie._sort_formats(info_dict['formats'])
        hdl.process_ie_result(info_dict)
        downloaded_ids = [info['format_id'] for info in hdl.downloaded_info_dicts]
        self.assertEqual(downloaded_ids, ['137', '141'])

        info_dict = _make_result(list(formats_order), extractor='youtube')
        hdl = HDL({'format': '(bestvideo[ext=mp4],bestvideo[ext=webm])+bestaudio'})
        yie = YoutubeIE(hdl)
        yie._sort_formats(info_dict['formats'])
        hdl.process_ie_result(info_dict)
        downloaded_ids = [info['format_id'] for info in hdl.downloaded_info_dicts]
        self.assertEqual(downloaded_ids, ['137+141', '248+141'])

        info_dict = _make_result(list(formats_order), extractor='youtube')
        hdl = HDL({'format': '(bestvideo[ext=mp4],bestvideo[ext=webm])[height<=720]+bestaudio'})
        yie = YoutubeIE(hdl)
        yie._sort_formats(info_dict['formats'])
        hdl.process_ie_result(info_dict)
        downloaded_ids = [info['format_id'] for info in hdl.downloaded_info_dicts]
        self.assertEqual(downloaded_ids, ['136+141', '247+141'])

        info_dict = _make_result(list(formats_order), extractor='youtube')
        hdl = HDL({'format': '(bestvideo[ext=none]/bestvideo[ext=webm])+bestaudio'})
        yie = YoutubeIE(hdl)
        yie._sort_formats(info_dict['formats'])
        hdl.process_ie_result(info_dict)
        downloaded_ids = [info['format_id'] for info in hdl.downloaded_info_dicts]
        self.assertEqual(downloaded_ids, ['248+141'])

        for f1, f2 in zip(formats_order, formats_order[1:]):
            info_dict = _make_result([f1, f2], extractor='youtube')
            hdl = HDL({'format': 'best/bestvideo'})
            yie = YoutubeIE(hdl)
            yie._sort_formats(info_dict['formats'])
            hdl.process_ie_result(info_dict)
            downloaded = hdl.downloaded_info_dicts[0]
            self.assertEqual(downloaded['format_id'], f1['format_id'])

            info_dict = _make_result([f2, f1], extractor='youtube')
            hdl = HDL({'format': 'best/bestvideo'})
            yie = YoutubeIE(hdl)
            yie._sort_formats(info_dict['formats'])
            hdl.process_ie_result(info_dict)
            downloaded = hdl.downloaded_info_dicts[0]
            self.assertEqual(downloaded['format_id'], f1['format_id'])

    def test_audio_only_extractor_format_selection(self):
        # For extractors with incomplete formats (all formats are audio-only or
        # video-only) best and worst should fallback to corresponding best/worst
        # video-only or audio-only formats (as per
        # https://github.com/ytdl-org/haruhi-dl/pull/5556)
        formats = [
            {'format_id': 'low', 'ext': 'mp3', 'preference': 1, 'vcodec': 'none', 'url': TEST_URL},
            {'format_id': 'high', 'ext': 'mp3', 'preference': 2, 'vcodec': 'none', 'url': TEST_URL},
        ]
        info_dict = _make_result(formats)

        hdl = HDL({'format': 'best'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'high')

        hdl = HDL({'format': 'worst'})
        hdl.process_ie_result(info_dict.copy())
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'low')

    def test_format_not_available(self):
        formats = [
            {'format_id': 'regular', 'ext': 'mp4', 'height': 360, 'url': TEST_URL},
            {'format_id': 'video', 'ext': 'mp4', 'height': 720, 'acodec': 'none', 'url': TEST_URL},
        ]
        info_dict = _make_result(formats)

        # This must fail since complete video-audio format does not match filter
        # and extractor does not provide incomplete only formats (i.e. only
        # video-only or audio-only).
        hdl = HDL({'format': 'best[height>360]'})
        self.assertRaises(ExtractorError, hdl.process_ie_result, info_dict.copy())

    def test_format_selection_issue_10083(self):
        # See https://github.com/ytdl-org/haruhi-dl/issues/10083
        formats = [
            {'format_id': 'regular', 'height': 360, 'url': TEST_URL},
            {'format_id': 'video', 'height': 720, 'acodec': 'none', 'url': TEST_URL},
            {'format_id': 'audio', 'vcodec': 'none', 'url': TEST_URL},
        ]
        info_dict = _make_result(formats)

        hdl = HDL({'format': 'best[height>360]/bestvideo[height>360]+bestaudio'})
        hdl.process_ie_result(info_dict.copy())
        self.assertEqual(hdl.downloaded_info_dicts[0]['format_id'], 'video+audio')

    def test_invalid_format_specs(self):
        def assert_syntax_error(format_spec):
            hdl = HDL({'format': format_spec})
            info_dict = _make_result([{'format_id': 'foo', 'url': TEST_URL}])
            self.assertRaises(SyntaxError, hdl.process_ie_result, info_dict)

        assert_syntax_error('bestvideo,,best')
        assert_syntax_error('+bestaudio')
        assert_syntax_error('bestvideo+')
        assert_syntax_error('/')

    def test_format_filtering(self):
        formats = [
            {'format_id': 'A', 'filesize': 500, 'width': 1000},
            {'format_id': 'B', 'filesize': 1000, 'width': 500},
            {'format_id': 'C', 'filesize': 1000, 'width': 400},
            {'format_id': 'D', 'filesize': 2000, 'width': 600},
            {'format_id': 'E', 'filesize': 3000},
            {'format_id': 'F'},
            {'format_id': 'G', 'filesize': 1000000},
        ]
        for f in formats:
            f['url'] = 'http://_/'
            f['ext'] = 'unknown'
        info_dict = _make_result(formats)

        hdl = HDL({'format': 'best[filesize<3000]'})
        hdl.process_ie_result(info_dict)
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'D')

        hdl = HDL({'format': 'best[filesize<=3000]'})
        hdl.process_ie_result(info_dict)
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'E')

        hdl = HDL({'format': 'best[filesize <= ? 3000]'})
        hdl.process_ie_result(info_dict)
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'F')

        hdl = HDL({'format': 'best [filesize = 1000] [width>450]'})
        hdl.process_ie_result(info_dict)
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'B')

        hdl = HDL({'format': 'best [filesize = 1000] [width!=450]'})
        hdl.process_ie_result(info_dict)
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'C')

        hdl = HDL({'format': '[filesize>?1]'})
        hdl.process_ie_result(info_dict)
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'G')

        hdl = HDL({'format': '[filesize<1M]'})
        hdl.process_ie_result(info_dict)
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'E')

        hdl = HDL({'format': '[filesize<1MiB]'})
        hdl.process_ie_result(info_dict)
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['format_id'], 'G')

        hdl = HDL({'format': 'all[width>=400][width<=600]'})
        hdl.process_ie_result(info_dict)
        downloaded_ids = [info['format_id'] for info in hdl.downloaded_info_dicts]
        self.assertEqual(downloaded_ids, ['B', 'C', 'D'])

        hdl = HDL({'format': 'best[height<40]'})
        try:
            hdl.process_ie_result(info_dict)
        except ExtractorError:
            pass
        self.assertEqual(hdl.downloaded_info_dicts, [])

    def test_default_format_spec(self):
        hdl = HDL({'simulate': True})
        self.assertEqual(hdl._default_format_spec({}), 'bestvideo+bestaudio/best')

        hdl = HDL({})
        self.assertEqual(hdl._default_format_spec({'is_live': True}), 'best/bestvideo+bestaudio')

        hdl = HDL({'simulate': True})
        self.assertEqual(hdl._default_format_spec({'is_live': True}), 'bestvideo+bestaudio/best')

        hdl = HDL({'outtmpl': '-'})
        self.assertEqual(hdl._default_format_spec({}), 'best/bestvideo+bestaudio')

        hdl = HDL({})
        self.assertEqual(hdl._default_format_spec({}, download=False), 'bestvideo+bestaudio/best')
        self.assertEqual(hdl._default_format_spec({'is_live': True}), 'best/bestvideo+bestaudio')


class TestHaruhiDL(unittest.TestCase):
    def test_subtitles(self):
        def s_formats(lang, autocaption=False):
            return [{
                'ext': ext,
                'url': 'http://localhost/video.%s.%s' % (lang, ext),
                '_auto': autocaption,
            } for ext in ['vtt', 'srt', 'ass']]
        subtitles = dict((l, s_formats(l)) for l in ['en', 'fr', 'es'])
        auto_captions = dict((l, s_formats(l, True)) for l in ['it', 'pt', 'es'])
        info_dict = {
            'id': 'test',
            'title': 'Test',
            'url': 'http://localhost/video.mp4',
            'subtitles': subtitles,
            'automatic_captions': auto_captions,
            'extractor': 'TEST',
        }

        def get_info(params={}):
            params.setdefault('simulate', True)
            hdl = HDL(params)
            hdl.report_warning = lambda *args, **kargs: None
            return hdl.process_video_result(info_dict, download=False)

        result = get_info()
        self.assertFalse(result.get('requested_subtitles'))
        self.assertEqual(result['subtitles'], subtitles)
        self.assertEqual(result['automatic_captions'], auto_captions)

        result = get_info({'writesubtitles': True})
        subs = result['requested_subtitles']
        self.assertTrue(subs)
        self.assertEqual(set(subs.keys()), set(['en']))
        self.assertTrue(subs['en'].get('data') is None)
        self.assertEqual(subs['en']['ext'], 'ass')

        result = get_info({'writesubtitles': True, 'subtitlesformat': 'foo/srt'})
        subs = result['requested_subtitles']
        self.assertEqual(subs['en']['ext'], 'srt')

        result = get_info({'writesubtitles': True, 'subtitleslangs': ['es', 'fr', 'it']})
        subs = result['requested_subtitles']
        self.assertTrue(subs)
        self.assertEqual(set(subs.keys()), set(['es', 'fr']))

        result = get_info({'writesubtitles': True, 'writeautomaticsub': True, 'subtitleslangs': ['es', 'pt']})
        subs = result['requested_subtitles']
        self.assertTrue(subs)
        self.assertEqual(set(subs.keys()), set(['es', 'pt']))
        self.assertFalse(subs['es']['_auto'])
        self.assertTrue(subs['pt']['_auto'])

        result = get_info({'writeautomaticsub': True, 'subtitleslangs': ['es', 'pt']})
        subs = result['requested_subtitles']
        self.assertTrue(subs)
        self.assertEqual(set(subs.keys()), set(['es', 'pt']))
        self.assertTrue(subs['es']['_auto'])
        self.assertTrue(subs['pt']['_auto'])

    def test_add_extra_info(self):
        test_dict = {
            'extractor': 'Foo',
        }
        extra_info = {
            'extractor': 'Bar',
            'playlist': 'funny videos',
        }
        HDL.add_extra_info(test_dict, extra_info)
        self.assertEqual(test_dict['extractor'], 'Foo')
        self.assertEqual(test_dict['playlist'], 'funny videos')

    def test_prepare_filename(self):
        info = {
            'id': '1234',
            'ext': 'mp4',
            'width': None,
            'height': 1080,
            'title1': '$PATH',
            'title2': '%PATH%',
        }

        def fname(templ):
            hdl = HaruhiDL({'outtmpl': templ})
            return hdl.prepare_filename(info)
        self.assertEqual(fname('%(id)s.%(ext)s'), '1234.mp4')
        self.assertEqual(fname('%(id)s-%(width)s.%(ext)s'), '1234-NA.mp4')
        # Replace missing fields with 'NA'
        self.assertEqual(fname('%(uploader_date)s-%(id)s.%(ext)s'), 'NA-1234.mp4')
        self.assertEqual(fname('%(height)d.%(ext)s'), '1080.mp4')
        self.assertEqual(fname('%(height)6d.%(ext)s'), '  1080.mp4')
        self.assertEqual(fname('%(height)-6d.%(ext)s'), '1080  .mp4')
        self.assertEqual(fname('%(height)06d.%(ext)s'), '001080.mp4')
        self.assertEqual(fname('%(height) 06d.%(ext)s'), ' 01080.mp4')
        self.assertEqual(fname('%(height)   06d.%(ext)s'), ' 01080.mp4')
        self.assertEqual(fname('%(height)0 6d.%(ext)s'), ' 01080.mp4')
        self.assertEqual(fname('%(height)0   6d.%(ext)s'), ' 01080.mp4')
        self.assertEqual(fname('%(height)   0   6d.%(ext)s'), ' 01080.mp4')
        self.assertEqual(fname('%%'), '%')
        self.assertEqual(fname('%%%%'), '%%')
        self.assertEqual(fname('%%(height)06d.%(ext)s'), '%(height)06d.mp4')
        self.assertEqual(fname('%(width)06d.%(ext)s'), 'NA.mp4')
        self.assertEqual(fname('%(width)06d.%%(ext)s'), 'NA.%(ext)s')
        self.assertEqual(fname('%%(width)06d.%(ext)s'), '%(width)06d.mp4')
        self.assertEqual(fname('Hello %(title1)s'), 'Hello $PATH')
        self.assertEqual(fname('Hello %(title2)s'), 'Hello %PATH%')

    def test_format_note(self):
        hdl = HaruhiDL()
        self.assertEqual(hdl._format_note({}), '')
        assertRegexpMatches(self, hdl._format_note({
            'vbr': 10,
        }), r'^\s*10k$')
        assertRegexpMatches(self, hdl._format_note({
            'fps': 30,
        }), r'^30fps$')

    def test_postprocessors(self):
        filename = 'post-processor-testfile.mp4'
        audiofile = filename + '.mp3'

        class SimplePP(PostProcessor):
            def run(self, info):
                with open(audiofile, 'wt') as f:
                    f.write('EXAMPLE')
                return [info['filepath']], info

        def run_pp(params, PP):
            with open(filename, 'wt') as f:
                f.write('EXAMPLE')
            hdl = HaruhiDL(params)
            hdl.add_post_processor(PP())
            hdl.post_process(filename, {'filepath': filename})

        run_pp({'keepvideo': True}, SimplePP)
        self.assertTrue(os.path.exists(filename), '%s doesn\'t exist' % filename)
        self.assertTrue(os.path.exists(audiofile), '%s doesn\'t exist' % audiofile)
        os.unlink(filename)
        os.unlink(audiofile)

        run_pp({'keepvideo': False}, SimplePP)
        self.assertFalse(os.path.exists(filename), '%s exists' % filename)
        self.assertTrue(os.path.exists(audiofile), '%s doesn\'t exist' % audiofile)
        os.unlink(audiofile)

        class ModifierPP(PostProcessor):
            def run(self, info):
                with open(info['filepath'], 'wt') as f:
                    f.write('MODIFIED')
                return [], info

        run_pp({'keepvideo': False}, ModifierPP)
        self.assertTrue(os.path.exists(filename), '%s doesn\'t exist' % filename)
        os.unlink(filename)

    def test_match_filter(self):
        class FilterHDL(HDL):
            def __init__(self, *args, **kwargs):
                super(FilterHDL, self).__init__(*args, **kwargs)
                self.params['simulate'] = True

            def process_info(self, info_dict):
                super(HDL, self).process_info(info_dict)

            def _match_entry(self, info_dict, incomplete):
                res = super(FilterHDL, self)._match_entry(info_dict, incomplete)
                if res is None:
                    self.downloaded_info_dicts.append(info_dict)
                return res

        first = {
            'id': '1',
            'url': TEST_URL,
            'title': 'one',
            'extractor': 'TEST',
            'duration': 30,
            'filesize': 10 * 1024,
            'playlist_id': '42',
            'uploader': "變態妍字幕版 太妍 тест",
            'creator': "тест ' 123 ' тест--",
        }
        second = {
            'id': '2',
            'url': TEST_URL,
            'title': 'two',
            'extractor': 'TEST',
            'duration': 10,
            'description': 'foo',
            'filesize': 5 * 1024,
            'playlist_id': '43',
            'uploader': "тест 123",
        }
        videos = [first, second]

        def get_videos(filter_=None):
            hdl = FilterHDL({'match_filter': filter_})
            for v in videos:
                hdl.process_ie_result(v, download=True)
            return [v['id'] for v in hdl.downloaded_info_dicts]

        res = get_videos()
        self.assertEqual(res, ['1', '2'])

        def f(v):
            if v['id'] == '1':
                return None
            else:
                return 'Video id is not 1'
        res = get_videos(f)
        self.assertEqual(res, ['1'])

        f = match_filter_func('duration < 30')
        res = get_videos(f)
        self.assertEqual(res, ['2'])

        f = match_filter_func('description = foo')
        res = get_videos(f)
        self.assertEqual(res, ['2'])

        f = match_filter_func('description =? foo')
        res = get_videos(f)
        self.assertEqual(res, ['1', '2'])

        f = match_filter_func('filesize > 5KiB')
        res = get_videos(f)
        self.assertEqual(res, ['1'])

        f = match_filter_func('playlist_id = 42')
        res = get_videos(f)
        self.assertEqual(res, ['1'])

        f = match_filter_func('uploader = "變態妍字幕版 太妍 тест"')
        res = get_videos(f)
        self.assertEqual(res, ['1'])

        f = match_filter_func('uploader != "變態妍字幕版 太妍 тест"')
        res = get_videos(f)
        self.assertEqual(res, ['2'])

        f = match_filter_func('creator = "тест \' 123 \' тест--"')
        res = get_videos(f)
        self.assertEqual(res, ['1'])

        f = match_filter_func("creator = 'тест \\' 123 \\' тест--'")
        res = get_videos(f)
        self.assertEqual(res, ['1'])

        f = match_filter_func(r"creator = 'тест \' 123 \' тест--' & duration > 30")
        res = get_videos(f)
        self.assertEqual(res, [])

    def test_playlist_items_selection(self):
        entries = [{
            'id': compat_str(i),
            'title': compat_str(i),
            'url': TEST_URL,
        } for i in range(1, 5)]
        playlist = {
            '_type': 'playlist',
            'id': 'test',
            'entries': entries,
            'extractor': 'test:playlist',
            'extractor_key': 'test:playlist',
            'webpage_url': 'http://example.com',
        }

        def get_downloaded_info_dicts(params):
            hdl = HDL(params)
            # make a deep copy because the dictionary and nested entries
            # can be modified
            hdl.process_ie_result(copy.deepcopy(playlist))
            return hdl.downloaded_info_dicts

        def get_ids(params):
            return [int(v['id']) for v in get_downloaded_info_dicts(params)]

        result = get_ids({})
        self.assertEqual(result, [1, 2, 3, 4])

        result = get_ids({'playlistend': 10})
        self.assertEqual(result, [1, 2, 3, 4])

        result = get_ids({'playlistend': 2})
        self.assertEqual(result, [1, 2])

        result = get_ids({'playliststart': 10})
        self.assertEqual(result, [])

        result = get_ids({'playliststart': 2})
        self.assertEqual(result, [2, 3, 4])

        result = get_ids({'playlist_items': '2-4'})
        self.assertEqual(result, [2, 3, 4])

        result = get_ids({'playlist_items': '2,4'})
        self.assertEqual(result, [2, 4])

        result = get_ids({'playlist_items': '10'})
        self.assertEqual(result, [])

        result = get_ids({'playlist_items': '3-10'})
        self.assertEqual(result, [3, 4])

        result = get_ids({'playlist_items': '2-4,3-4,3'})
        self.assertEqual(result, [2, 3, 4])

        # Tests for https://github.com/ytdl-org/haruhi-dl/issues/10591
        # @{
        result = get_downloaded_info_dicts({'playlist_items': '2-4,3-4,3'})
        self.assertEqual(result[0]['playlist_index'], 2)
        self.assertEqual(result[1]['playlist_index'], 3)

        result = get_downloaded_info_dicts({'playlist_items': '2-4,3-4,3'})
        self.assertEqual(result[0]['playlist_index'], 2)
        self.assertEqual(result[1]['playlist_index'], 3)
        self.assertEqual(result[2]['playlist_index'], 4)

        result = get_downloaded_info_dicts({'playlist_items': '4,2'})
        self.assertEqual(result[0]['playlist_index'], 4)
        self.assertEqual(result[1]['playlist_index'], 2)
        # @}

    def test_urlopen_no_file_protocol(self):
        # see https://github.com/ytdl-org/haruhi-dl/issues/8227
        hdl = HDL()
        self.assertRaises(compat_urllib_error.URLError, hdl.urlopen, 'file:///etc/passwd')

    def test_do_not_override_ie_key_in_url_transparent(self):
        hdl = HDL()

        class Foo1IE(InfoExtractor):
            _VALID_URL = r'foo1:'

            def _real_extract(self, url):
                return {
                    '_type': 'url_transparent',
                    'url': 'foo2:',
                    'ie_key': 'Foo2',
                    'title': 'foo1 title',
                    'id': 'foo1_id',
                }

        class Foo2IE(InfoExtractor):
            _VALID_URL = r'foo2:'

            def _real_extract(self, url):
                return {
                    '_type': 'url',
                    'url': 'foo3:',
                    'ie_key': 'Foo3',
                }

        class Foo3IE(InfoExtractor):
            _VALID_URL = r'foo3:'

            def _real_extract(self, url):
                return _make_result([{'url': TEST_URL}], title='foo3 title')

        hdl.add_info_extractor(Foo1IE(hdl))
        hdl.add_info_extractor(Foo2IE(hdl))
        hdl.add_info_extractor(Foo3IE(hdl))
        hdl.extract_info('foo1:')
        downloaded = hdl.downloaded_info_dicts[0]
        self.assertEqual(downloaded['url'], TEST_URL)
        self.assertEqual(downloaded['title'], 'foo1 title')
        self.assertEqual(downloaded['id'], 'testid')
        self.assertEqual(downloaded['extractor'], 'testex')
        self.assertEqual(downloaded['extractor_key'], 'TestEx')


if __name__ == '__main__':
    unittest.main()
